from contextlib import contextmanager
import gzip
import json
import os
import pytest
import pickle
import sys
import tempfile
import vcr

from gcsfs.core import GCSFileSystem
from gcsfs.tests.settings import (TEST_BUCKET, TEST_PROJECT, RECORD_MODE,
                                  GOOGLE_TOKEN)


def before_record_response(response):
    r = pickle.loads(pickle.dumps(response))
    try:
        try:
            data = json.loads(gzip.decompress(r['body']['string']).decode())
            if 'access_token' in data:
                data['access_token'] = 'xxx'
            if 'id_token' in data:
                data['id_token'] = 'xxx'
            r['body']['string'] = gzip.compress(
                    json.dumps(data).replace(
                            TEST_PROJECT, 'test_project').encode())
        except (OSError, TypeError, ValueError):
            r['body']['string'] = r['body']['string'].replace(
                    TEST_PROJECT.encode(), b'test_project')
    except Exception:
        pass
    return r


def before_record(request):
    r = pickle.loads(pickle.dumps(request))
    r.uri = request.uri.replace(TEST_PROJECT, 'test_project')
    return r


def matcher(r1, r2):
    if r1.uri.replace(TEST_PROJECT,
                      'test_project') != r2.uri.replace(TEST_PROJECT,
                      'test_project'):
        print('uri mismatch')
        return False
    if r1.body != r2.body:
        print('body mismatch')
        return False
    if r1.method != r2.method:
        print('method mismatch')
        return False
    for key in ['Content-Length', 'Content-Type', 'Range', 'Server']:
        if r1.head.get(key, '') != r2.head.get(key, ''):
            print('header mismatch', key)
            return False
    print('match')
    return True

recording_path = os.path.join(os.path.dirname(__file__), 'recordings')

my_vcr = vcr.VCR(
    cassette_library_dir=recording_path,
    record_mode=RECORD_MODE,
    path_transformer=vcr.VCR.ensure_suffix('.yaml'),
    filter_headers=['Authorization'],
    filter_query_parameters=['refresh_token', 'upload_id', 'client_id',
                             'client_secret'],
    before_record_response=before_record_response,
    before_record=before_record
    )
my_vcr.register_matcher('all', matcher)
files = {'test/accounts.1.json':  (b'{"amount": 100, "name": "Alice"}\n'
                                   b'{"amount": 200, "name": "Bob"}\n'
                                   b'{"amount": 300, "name": "Charlie"}\n'
                                   b'{"amount": 400, "name": "Dennis"}\n'),
         'test/accounts.2.json':  (b'{"amount": 500, "name": "Alice"}\n'
                                   b'{"amount": 600, "name": "Bob"}\n'
                                   b'{"amount": 700, "name": "Charlie"}\n'
                                   b'{"amount": 800, "name": "Dennis"}\n')}

csv_files = {'2014-01-01.csv': (b'name,amount,id\n'
                                b'Alice,100,1\n'
                                b'Bob,200,2\n'
                                b'Charlie,300,3\n'),
             '2014-01-02.csv': (b'name,amount,id\n'),
             '2014-01-03.csv': (b'name,amount,id\n'
                                b'Dennis,400,4\n'
                                b'Edith,500,5\n'
                                b'Frank,600,6\n')}
text_files = {'nested/file1': b'hello\n',
              'nested/file2': b'world',
              'nested/nested2/file1': b'hello\n',
              'nested/nested2/file2': b'world'}
a = TEST_BUCKET+'/tmp/test/a'
b = TEST_BUCKET+'/tmp/test/b'
c = TEST_BUCKET+'/tmp/test/c'
d = TEST_BUCKET+'/tmp/test/d'


@contextmanager
def ignoring(*exceptions):
    try:
        yield
    except exceptions:
        pass


@pytest.yield_fixture()
def tempdir():
    d = tempfile.mkdtemp()
    yield d
    if os.path.exists(d):
        shutil.rmtree(d, ignore_errors=True)


@contextmanager
def tmpfile(extension='', dir=None):
    extension = '.' + extension.lstrip('.')
    handle, filename = tempfile.mkstemp(extension, dir=dir)
    os.close(handle)
    os.remove(filename)

    try:
        yield filename
    finally:
        if os.path.exists(filename):
            if os.path.isdir(filename):
                shutil.rmtree(filename)
            else:
                with ignoring(OSError):
                    os.remove(filename)


@pytest.yield_fixture
def token_restore():
    try:
        cache = GCSFileSystem.tokens
        GCSFileSystem.tokens = {}
        yield
    finally:
        GCSFileSystem.tokens = cache
        GCSFileSystem._save_tokens()


@contextmanager
def gcs_maker(populate=False):
    gcs = GCSFileSystem(TEST_PROJECT, token=GOOGLE_TOKEN)
    try:
        if not gcs.exists(TEST_BUCKET):
            gcs.mkdir(TEST_BUCKET)
        for k in [a, b, c, d]:
            try:
                gcs.rm(k)
            except:
                pass
        if populate:
            for flist in [files, csv_files, text_files]:
                for fname, data in flist.items():
                    with gcs.open(TEST_BUCKET+'/'+fname, 'wb') as f:
                        f.write(data)
        yield gcs
    finally:
        [gcs.rm(f) for f in gcs.walk(TEST_BUCKET)]
        # gcs.rmdir(settings.TEST_BUCKET)



import pytest
from gcsfs.tests.settings import TEST_PROJECT, GOOGLE_TOKEN, TEST_BUCKET
from gcsfs.tests.utils import (tempdir, token_restore, my_vcr, gcs_maker,
                               files, csv_files, text_files, a, b, c, d)
from gcsfs import GCSMap, GCSFileSystem, core

root = TEST_BUCKET+'/mapping'

@my_vcr.use_cassette(match=['all'])
def test_map_simple(token_restore):
    with gcs_maker() as gcs:
        d = GCSMap(root, gcs)
        assert not d

        assert list(d) == list(d.keys()) == []
        assert list(d.values()) == []
        assert list(d.items()) == []
        d = GCSMap(root, gcs, check=True)


@my_vcr.use_cassette(match=['all'])
def test_map_default_gcsfilesystem(token_restore):
    with gcs_maker() as gcs:
        d = GCSMap(root)
        assert d.gcs is gcs


@my_vcr.use_cassette(match=['all'])
def test_map_errors(token_restore):
    with gcs_maker() as gcs:
        d = GCSMap(root, gcs)
        with pytest.raises(KeyError):
            d['nonexistent']
        try:
            GCSMap('does-not-exist')
        except Exception as e:
            assert 'does-not-exist' in str(e)


@my_vcr.use_cassette(match=['all'])
def test_map_with_data(token_restore):
    with gcs_maker() as gcs:
        d = GCSMap(root, gcs)
        d['x'] = b'123'
        assert list(d) == list(d.keys()) == ['x']
        assert list(d.values()) == [b'123']
        assert list(d.items()) == [('x', b'123')]
        assert d['x'] == b'123'
        assert bool(d)

        assert gcs.walk(root) == [TEST_BUCKET+'/mapping/x']
        d['x'] = b'000'
        assert d['x'] == b'000'

        d['y'] = b'456'
        assert d['y'] == b'456'
        assert set(d) == {'x', 'y'}

        d.clear()
        assert list(d) == []


@my_vcr.use_cassette(match=['all'])
def test_map_complex_keys(token_restore):
    with gcs_maker() as gcs:
        d = GCSMap(root, gcs)
        d[1] = b'hello'
        assert d[1] == b'hello'
        del d[1]

        d[1, 2] = b'world'
        assert d[1, 2] == b'world'
        del d[1, 2]

        d['x', 1, 2] = b'hello world'
        assert d['x', 1, 2] == b'hello world'

        assert ('x', 1, 2) in d


@my_vcr.use_cassette(match=['all'])
def test_map_clear_empty(token_restore):
    with gcs_maker() as gcs:
        d = GCSMap(root, gcs)
        d.clear()
        assert list(d) == []
        d[1] = b'1'
        assert list(d) == ['1']
        d.clear()
        assert list(d) == []


@my_vcr.use_cassette(match=['all'])
def test_map_pickle(token_restore):
    with gcs_maker() as gcs:
        d = GCSMap(root, gcs)
        d['x'] = b'1'
        assert d['x'] == b'1'

        import pickle
        d2 = pickle.loads(pickle.dumps(d))

        assert d2['x'] == b'1'


@my_vcr.use_cassette(match=['all'])
def test_map_array(token_restore):
    with gcs_maker() as gcs:
        from array import array
        d = GCSMap(root, gcs)
        d['x'] = array('B', [65] * 1000)

        assert d['x'] == b'A' * 1000


@my_vcr.use_cassette(match=['all'])
def test_map_bytearray(token_restore):
    with gcs_maker() as gcs:
        from array import array
        d = GCSMap(root, gcs)
        d['x'] = bytearray(b'123')

        assert d['x'] == b'123'


@my_vcr.use_cassette(match=['all'])
def test_new_bucket(token_restore):
    with gcs_maker() as gcs:
        new_bucket = TEST_BUCKET + 'new-bucket'
        try:
            gcs.rmdir(new_bucket)
        except:
            pass
        with pytest.raises(ValueError) as e:
            d = GCSMap(new_bucket, gcs)
        assert 'create=True' in str(e)

        try:
            d = GCSMap(new_bucket, gcs, create=True)
            assert not d

            d = GCSMap(new_bucket + '/new-directory', gcs)
            assert not d
        finally:
            gcs.rmdir(new_bucket)



# -*- coding: utf-8 -*-
"""
Google Cloud Storage pythonic interface
"""
from __future__ import print_function

import array
import io
import json
import logging
import oauth2client.client   # version 1.5.2
import os
import pickle
import re
import requests
import sys
import time
import warnings
import webbrowser

from .utils import read_block
PY2 = sys.version_info.major == 2

logger = logging.getLogger(__name__)

# client created 23-Sept-2017
not_secret = {"client_id": "586241054156-7a3vrghs70ffkkfkmnnatjnbjg03cq9a."
                           "apps.googleusercontent.com",
              "client_secret": "RsSQZJKYE00oRv2ibYlj28Sx"}
tfile = os.path.join(os.path.expanduser("~"), '.gcs_tokens')
ACLs = {"authenticatedread", "bucketownerfullcontrol", "bucketownerread",
        "private", "projectprivate", "publicread"}
bACLs = {"authenticatedRead", "private", "projectPrivate", "publicRead",
         "publicReadWrite"}
DEFAULT_PROJECT = os.environ.get('GCSFS_DEFAULT_PROJECT', '')
DEBUG = False

if PY2:
    FileNotFoundError = IOError


def quote_plus(s):
    """
    Convert some URL elements to be HTTP-safe.

    Not the same as in urllib, because, for instance, parentheses and commas
    are passed through.

    Parameters
    ----------
    s: input URL/portion

    Returns
    -------
    corrected URL
    """
    s = s.replace('/', '%2F')
    s = s.replace(' ', '%20')
    return s


def split_path(path):
    """
    Normalise GCS path string into bucket and key.

    Parameters
    ----------
    path : string
        Input path, like `gcs://mybucket/path/to/file`

    Examples
    --------
    >>> split_path("gcs://mybucket/path/to/file")
    ['mybucket', 'path/to/file']
    """
    if path.startswith('gcs://'):
        path = path[6:]
    if path.startswith('gs://'):
        path = path[5:]
    if '/' not in path:
        return path, ""
    else:
        return path.split('/', 1)


def validate_response(r, path):
    """
    Check the requests object r, raise error if it's not ok.

    Parameters
    ----------
    r: requests response object
    path: associated URL path, for error messages
    """
    if not r.ok:
        m = str(r.content)
        try:
            msg = r.json()['error']['message']
        except:
            msg = str(r.content)

        if DEBUG:
            print(r.url, r.headers, sep='\n')
        if "Not Found" in m:
            raise FileNotFoundError(path)
        elif "forbidden" in m:
            raise IOError("Forbidden: %s\n%s" % (path, msg))
        elif "invalid" in m:
            raise ValueError("Bad Request: %s\n%s" % (path, msg))
        else:
            raise RuntimeError(m)


class GCSFileSystem(object):
    """
    Connect to Google Cloud Storage.

    Two modes of authentication are supported:
    - if ``token=None``, you will be given a "device code", which you must
      enter into a browser where you are logged in with your Google identity.
    - if ``token='cloud'``, we assume we are running within google compute
      or google container engine, and query the internal metadata directly for
      a token.
    - you may supply a token generated by the
      [gcloud](https://cloud.google.com/sdk/docs/)
      utility; this is either a python dictionary, or the name of a file
      containing the JSON returned by logging in with the gcloud CLI tool. On
      a posix system this may be at
      ``~/.config/gcloud/application_default_credentials.json``

    We maintain a cache of refresh tokens in the file ~/.gcs_tokens, so for any
    pair of (project, access), you will not need to log in once your credentials
    are verified.

    Parameters
    ----------
    project : string
        GCS users may only access to contents of one project in a single
        instance of GCSFileSystem. This is required in order
        to list all the buckets you have access to within a project.
    access : one of {'read_only', 'read_write', 'full_control'}
        Full control implies read/write as well as modifying metadata,
        e.g., access control.
    token: None, dict or string
        (see description of authentication methods, above)
    """
    scopes = {'read_only', 'read_write', 'full_control'}
    retries = 10
    base = "https://www.googleapis.com/storage/v1/"
    _singleton = [None]
    default_block_size = 5 * 2**20

    def __init__(self, project=DEFAULT_PROJECT, access='full_control',
                 token=None, block_size=None):
        if access not in self.scopes:
            raise ValueError('access must be one of {}', self.scopes)
        if project is None:
            warnings.warn('GCS project not set - cannot list or create buckets')
        self.input_token = token
        if block_size is not None:
            self.default_block_size = block_size
        self.project = project
        self.access = access
        self.dirs = {}
        self.connect()
        self._singleton[0] = self

    @classmethod
    def current(cls):
        """ Return the most recently created GCSFileSystem

        If no GCSFileSystem has been created, then create one
        """
        if not cls._singleton[0]:
            return GCSFileSystem()
        else:
            return cls._singleton[0]

    @staticmethod
    def _parse_gtoken(gt):
        if isinstance(gt, str):
            t = json.load(open(gt))
        else:
            t = gt.copy()
        typ = t.pop('type')
        if typ != "authorized_user":
            raise ValueError("Only 'authorized_user' tokens accepted, got: %s"
                             % typ)
        t['grant_type'] = 'refresh_token'
        t['timestamp'] = time.time()
        t['expires_in'] = 0
        return t

    @staticmethod
    def load_tokens():
        try:
            with open(tfile, 'rb') as f:
                tokens = pickle.load(f)
        except Exception:
            tokens = {}
        GCSFileSystem.tokens = tokens

    def connect(self, refresh=False):
        """
        Establish session token. A new token will be requested if the current
        one is within 100s of expiry.

        Parameters
        ----------
        refresh: bool (False)
            Force refresh, even if the token is expired.
        """
        token = self.input_token
        project, access = self.project, self.access
        if token is not None and str(token) != 'cloud':
            if 'type' in token or isinstance(token, str):
                token = self._parse_gtoken(token)
            self.tokens[(project, access)] = token
        if (project, access) in self.tokens:
            # cached credentials
            data = self.tokens[(project, access)]
        elif token == 'cloud':
            data = {'timestamp': time.time() - 3600, 'expires_in': 1,
                    'type': 'cloud'}
        else:
            # no credentials - try to ask google in the browser
            scope = "https://www.googleapis.com/auth/devstorage." + access
            path = 'https://accounts.google.com/o/oauth2/device/code'
            r = requests.post(path,
                              params={'client_id': not_secret['client_id'],
                                      'scope': scope})
            validate_response(r, path)
            data = json.loads(r.content.decode())
            print('Enter the following code when prompted in the browser:')
            print(data['user_code'])
            webbrowser.open(data['verification_url'])
            for i in range(self.retries):
                time.sleep(2)
                r = requests.post(
                        "https://www.googleapis.com/oauth2/v4/token",
                        params={'client_id': not_secret['client_id'],
                                'client_secret': not_secret['client_secret'],
                                'code': data['device_code'],
                                'grant_type':
                                    "http://oauth.net/grant_type/device/1.0"})
                data2 = json.loads(r.content.decode())
                if 'error' in data2:
                    if i == self.retries - 1:
                        raise RuntimeError("Waited too long for browser"
                                           "authentication.")
                    continue
                data = data2
                break
            data['timestamp'] = time.time()
            data.update(not_secret)
        if refresh or time.time() - data['timestamp'] > data['expires_in'] - 100:
            # token has expired, or is about to - call refresh
            if data.get('type', None) == 'cloud':
                path = ('http://metadata.google.internal/computeMetadata/v1/'
                        'instance/service-accounts/default/token')
                r = requests.get(path, headers={'Metadata-Flavor': 'Google'})
                validate_response(r, path)
                data = r.json()
                data['timestamp'] = time.time()
                data['type'] = 'cloud'
            else:
                path = "https://www.googleapis.com/oauth2/v4/token"
                r = requests.post(
                        path,
                        params={'client_id': data['client_id'],
                                'client_secret': data['client_secret'],
                                'refresh_token': data['refresh_token'],
                                'grant_type': "refresh_token"})
                validate_response(r, path)
                data['timestamp'] = time.time()
                data['access_token'] = r.json()['access_token']

        self.tokens[(project, access)] = data
        self.header = {'Authorization': 'Bearer ' + data['access_token']}
        self._save_tokens()

    @staticmethod
    def _save_tokens():
        try:
            with open(tfile, 'wb') as f:
                pickle.dump(GCSFileSystem.tokens, f, 2)
        except Exception as e:
            warnings.warn('Saving token cache failed: ' + str(e))

    def _call(self, method, path, *args, **kwargs):
        for k, v in list(kwargs.items()):
            # only pass parameters that have values
            if v is None:
                del kwargs[k]
        json = kwargs.pop('json', None)
        meth = getattr(requests, method)
        if args:
            path = path.format(*[quote_plus(p) for p in args])
        r = meth(self.base + path, headers=self.header, params=kwargs,
                 json=json)
        try:
            out = r.json()
        except ValueError:
            out = r.content
        validate_response(r, path)
        return out

    def _list_buckets(self):
        if '' not in self.dirs:
            out = self._call('get', 'b/', project=self.project)
            dirs = out.get('items', [])
            self.dirs[''] = dirs
        return self.dirs['']

    def _list_bucket(self, bucket, max_results=1000):
        if bucket not in self.dirs:
            out = self._call('get', 'b/{}/o/', bucket, maxResults=max_results)
            dirs = out.get('items', [])
            next_page_token = out.get('nextPageToken', None)
            while next_page_token is not None:
                out = self._call('get', 'b/{}/o/', bucket, maxResults=max_results, pageToken=next_page_token)
                dirs.extend(out.get('items', []))
                next_page_token = out.get('nextPageToken', None)
            for f in dirs:
                f['name'] = '%s/%s' % (bucket, f['name'])
                f['size'] = int(f.get('size'), 0)
            self.dirs[bucket] = dirs
        return self.dirs[bucket]

    def mkdir(self, bucket, acl='projectPrivate',
              default_acl='bucketOwnerFullControl'):
        """
        New bucket

        Parameters
        ----------
        bucket: str
            bucket name
        acl: string, one of bACLs
            access for the bucket itself
        default_acl: str, one of ACLs
            default ACL for objects created in this bucket
        """
        self._call('post', 'b/', predefinedAcl=acl, project=self.project,
                   predefinedDefaultObjectAcl=default_acl,
                   json={"name": bucket})
        self.invalidate_cache(bucket)
        self.invalidate_cache('')

    def rmdir(self, bucket):
        """Delete an empty bucket"""
        self._call('delete', 'b/' + bucket)
        if '' in self.dirs:
            for v in self.dirs[''][:]:
                if v['name'] == bucket:
                    self.dirs[''].remove(v)
        self.invalidate_cache(bucket)

    def invalidate_cache(self, bucket=None):
        """
        Mark files cache as dirty, so that it is reloaded on next use.

        Parameters
        ----------
        bucket: string or None
            If None, clear all files cached; if a string, clear the files
            corresponding to that bucket.
        """
        if bucket in {'/', '', None}:
            self.dirs.clear()
        else:
            self.dirs.pop(bucket, None)

    def ls(self, path, detail=False):
        if path in ['', '/']:
            out = self._list_buckets()
        else:
            bucket, prefix = split_path(path)
            path = '/'.join([bucket, prefix])
            files = self._list_bucket(bucket)
            seek, l, bit = (path, len(path), '') if path.endswith('/') else (
                path+'/', len(path)+1, '/')
            out = []
            for f in files:
                if (f['name'].startswith(seek) and '/' not in f['name'][l:] or
                        f['name'] == path):
                    out.append(f)
                elif f['name'].startswith(seek) and '/' in f['name'][l:]:
                    directory = {
                        'bucket': bucket, 'kind': 'storage#object',
                        'size': 0, 'storageClass': 'DIRECTORY',
                        'name': path+bit+f['name'][l:].split('/', 1)[0]+'/'}
                    if directory not in out:
                        out.append(directory)
        if detail:
            return out
        else:
            return [f['name'] for f in out]

    def walk(self, path, detail=False):
        bucket, prefix = split_path(path)
        if not bucket:
            raise ValueError('Cannot walk all of GCS')
        path = '/'.join([bucket, prefix])
        files = self._list_bucket(bucket)
        if path.endswith('/'):
            files = [f for f in files if f['name'].startswith(path) or
                     f['name'] == path]
        else:
            files = [f for f in files if f['name'].startswith(path+'/') or
                     f['name'] == path]
        if detail:
            return files
        else:
            return [f['name'] for f in files]

    def du(self, path, total=False, deep=False):
        if deep:
            files = self.walk(path, True)
        else:
            files = [f for f in self.ls(path, True)]
        if total:
            return sum(f['size'] for f in files)
        return {f['name']: f['size'] for f in files}

    def glob(self, path):
        """
        Find files by glob-matching.

        Note that the bucket part of the path must not contain a "*"
        """
        path = path.rstrip('/')
        bucket, key = split_path(path)
        path = '/'.join([bucket, key])
        if "*" in bucket:
            raise ValueError('Bucket cannot contain a "*"')
        if '*' not in path:
            path = path.rstrip('/') + '/*'
        if '/' in path[:path.index('*')]:
            ind = path[:path.index('*')].rindex('/')
            root = path[:ind + 1]
        else:
            root = ''
        allfiles = self.walk(root)
        pattern = re.compile("^" + path.replace('//', '/')
                             .rstrip('/')
                             .replace('*', '[^/]*')
                             .replace('?', '.') + "$")
        out = [f for f in allfiles if re.match(pattern,
               f.replace('//', '/').rstrip('/'))]
        return out

    def exists(self, path):
        bucket, key = split_path(path)
        try:
            if key:
                return bool(self.info(path))
            else:
                return bucket in self.ls('')
        except FileNotFoundError:
            return False

    def info(self, path):
        path = '/'.join(split_path(path))
        files = self.ls(path, True)
        out = [f for f in files if f['name'] == path]
        if out:
            return out[0]
        else:
            raise FileNotFoundError(path)

    def url(self, path):
        return self.info(path)['mediaLink']

    def cat(self, path):
        """ Simple one-shot get of file data """
        details = self.info(path)
        return _fetch_range(self.header, details)

    def get(self, rpath, lpath, blocksize=5 * 2 ** 20):
        with self.open(rpath, 'rb', block_size=blocksize) as f1:
            with open(lpath, 'wb') as f2:
                while True:
                    d = f1.read(blocksize)
                    if not d:
                        break
                    f2.write(d)

    def put(self, lpath, rpath, blocksize=5 * 2 ** 20, acl=None):
        with self.open(rpath, 'wb', block_size=blocksize, acl=acl) as f1:
            with open(lpath, 'rb') as f2:
                while True:
                    d = f2.read(blocksize)
                    if not d:
                        break
                    f1.write(d)

    def head(self, path, size=1024):
        with self.open(path, 'rb') as f:
            return f.read(size)

    def tail(self, path, size=1024):
        if size > self.info(path)['size']:
            return self.cat(path)
        with self.open(path, 'rb') as f:
            f.seek(-size, 2)
            return f.read()

    def merge(self, path, paths, acl=None):
        """Concatenate objects within a single bucket"""
        bucket, key = split_path(path)
        source = [{'name': split_path(p)[1]} for p in paths]
        self._call('post', 'b/{}/o/{}/compose', bucket, key,
                   destinationPredefinedAcl=acl,
                   json={'sourceObjects': source,
                         "kind": "storage#composeRequest",
                         'destination': {'name': key, 'bucket': bucket}})

    def copy(self, path1, path2, acl=None):
        b1, k1 = split_path(path1)
        b2, k2 = split_path(path2)
        self._call('post', 'b/{}/o/{}/copyTo/b/{}/o/{}', b1, k1, b2, k2,
                   destinationPredefinedAcl=acl)

    def mv(self, path1, path2, acl=None):
        self.copy(path1, path2, acl)
        self.rm(path1)

    def rm(self, path):
        bucket, path = split_path(path)
        self._call('delete', "b/{}/o/{}", bucket, path)
        self.invalidate_cache(bucket)

    def open(self, path, mode='rb', block_size=None, acl=None):
        if block_size is None:
            block_size = self.default_block_size
        return GCSFile(self, path, mode, block_size)

    def touch(self, path):
        with self.open(path, 'wb'):
            pass

    def read_block(self, fn, offset, length, delimiter=None):
        """ Read a block of bytes from a GCS file

        Starting at ``offset`` of the file, read ``length`` bytes.  If
        ``delimiter`` is set then we ensure that the read starts and stops at
        delimiter boundaries that follow the locations ``offset`` and ``offset
        + length``.  If ``offset`` is zero then we start at zero.  The
        bytestring returned WILL include the end delimiter string.

        If offset+length is beyond the eof, reads to eof.

        Parameters
        ----------
        fn: string
            Path to filename on GCS
        offset: int
            Byte offset to start read
        length: int
            Number of bytes to read
        delimiter: bytes (optional)
            Ensure reading starts and stops at delimiter bytestring

        Examples
        --------
        >>> gcs.read_block('data/file.csv', 0, 13)  # doctest: +SKIP
        b'Alice, 100\\nBo'
        >>> gcs.read_block('data/file.csv', 0, 13, delimiter=b'\\n')  # doctest: +SKIP
        b'Alice, 100\\nBob, 200\\n'

        Use ``length=None`` to read to the end of the file.
        >>> gcs.read_block('data/file.csv', 0, None, delimiter=b'\\n')  # doctest: +SKIP
        b'Alice, 100\\nBob, 200\\nCharlie, 300'

        See Also
        --------
        distributed.utils.read_block
        """
        with self.open(fn, 'rb') as f:
            size = f.size
            if length is None:
                length = size
            if offset + length > size:
                length = size - offset
            bytes = read_block(f, offset, length, delimiter)
        return bytes

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['header']
        logger.debug("Serialize with state: %s", d)
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.connect()


GCSFileSystem.load_tokens()


class GCSFile:

    def __init__(self, gcsfs, path, mode='rb', block_size=5 * 2 ** 20,
                 acl=None):
        bucket, key = split_path(path)
        self.gcsfs = gcsfs
        self.bucket = bucket
        self.key = key
        self.mode = mode
        self.blocksize = block_size
        self.cache = b""
        self.loc = 0
        self.acl = acl
        self.end = None
        self.start = None
        self.closed = False
        self.trim = True
        if mode not in {'rb', 'wb'}:
            raise NotImplementedError('File mode not supported')
        if mode == 'rb':
            self.details = gcsfs.info(path)
            self.size = self.details['size']
        else:
            if block_size < 2**18:
                warnings.warn('Setting block size to minimum value, 2**18')
                self.blocksize = 2**18
            self.buffer = io.BytesIO()
            self.offset = 0
            self.forced = False

    def info(self):
        """ File information about this path """
        return self.details

    def url(self):
        return self.details['mediaLink']

    def tell(self):
        """ Current file location """
        return self.loc

    def seek(self, loc, whence=0):
        """ Set current file location

        Parameters
        ----------
        loc : int
            byte location
        whence : {0, 1, 2}
            from start of file, current location or end of file, resp.
        """
        if not self.mode == 'rb':
            raise ValueError('Seek only available in read mode')
        if whence == 0:
            nloc = loc
        elif whence == 1:
            nloc = self.loc + loc
        elif whence == 2:
            nloc = self.size + loc
        else:
            raise ValueError(
                "invalid whence (%s, should be 0, 1 or 2)" % whence)
        if nloc < 0:
            raise ValueError('Seek before start of file')
        self.loc = nloc
        return self.loc

    def readline(self, length=-1):
        """
        Read and return a line from the stream.

        If length is specified, at most size bytes will be read.
        """
        self._fetch(self.loc, self.loc + 1)
        while True:
            found = self.cache[self.loc - self.start:].find(b'\n') + 1
            if 0 < length < found:
                return self.read(length)
            if found:
                return self.read(found)
            if self.end > self.size:
                return self.read(length)
            self._fetch(self.start, self.end + self.blocksize)

    def __next__(self):
        data = self.readline()
        if data:
            return data
        else:
            raise StopIteration

    next = __next__

    def __iter__(self):
        return self

    def readlines(self):
        """ Return all lines in a file as a list """
        return list(self)

    def write(self, data):
        """
        Write data to buffer.

        Buffer only sent to GCS on flush() or if buffer is greater than
        or equal to blocksize.

        Parameters
        ----------
        data : bytes
            Set of bytes to be written.
        """
        if self.mode not in {'wb', 'ab'}:
            raise ValueError('File not in write mode')
        if self.closed:
            raise ValueError('I/O operation on closed file.')
        out = self.buffer.write(ensure_writable(data))
        self.loc += out
        if self.buffer.tell() >= self.blocksize:
            self.flush()
        return out

    def flush(self, force=False):
        """
        Write buffered data to GCS.

        Uploads the current buffer, if it is larger than the block-size, or if
        the file is being closed.

        Parameters
        ----------
        force : bool
            When closing, write the last block even if it is smaller than
            blocks are allowed to be.
        """
        if self.mode not in {'wb', 'ab'}:
            raise ValueError('Flush on a file not in write mode')
        if self.closed:
            raise ValueError('Flush on closed file')
        if self.buffer.tell() == 0 and not force:
            # no data in the buffer to write
            return
        if force and self.forced:
            raise ValueError("Force flush cannot be called more than once")
        if force:
            self.forced = True
        if force and not self.offset and self.buffer.tell() <= 5 * 2**20:
            self._simple_upload()
            return

        if not self.offset:
            self._initiate_upload()
        self._upload_chunk(final=force)

    def _upload_chunk(self, final=False):
        self.buffer.seek(0)
        data = self.buffer.read()
        head = self.gcsfs.header.copy()
        l = len(data)
        if final:
            if l:
                head['Content-Range'] = 'bytes %i-%i/%i' % (
                    self.offset, self.offset + l - 1, self.offset + l)
            else:
                # closing when buffer is empty
                head['Content-Range'] = 'bytes */%i' % self.offset
                data = None
        else:
            head['Content-Range'] = 'bytes %i-%i/*' % (
                self.offset, self.offset + l - 1)
        head.update({'Content-Type': 'application/octet-stream',
                     'Content-Length': str(l)})
        r = requests.post(self.location, params={'uploadType': 'resumable'},
                          headers=head, data=data)
        validate_response(r, self.location)
        if 'Range' in r.headers:
            shortfall = (self.offset + l - 1) - int(
                    r.headers['Range'].split('-')[1])
            if shortfall:
                self.buffer = io.BytesIO(data[-shortfall:])
                self.buffer.seek(shortfall)
            else:
                self.buffer = io.BytesIO()
            import pdb
            # pdb.set_trace()
            self.offset += l - shortfall
        else:
            self.buffer = io.BytesIO()
            self.offset += l

    def _initiate_upload(self):
        r = requests.post('https://www.googleapis.com/upload/storage/v1/b/%s/o'
                          % quote_plus(self.bucket),
                          params={'uploadType': 'resumable'},
                          headers=self.gcsfs.header, json={'name': self.key})
        self.location = r.headers['Location']

    def _simple_upload(self):
        """One-shot upload, less than 5MB"""
        head = self.gcsfs.header.copy()
        self.buffer.seek(0)
        data = self.buffer.read()
        path = ('https://www.googleapis.com/upload/storage/v1/b/%s/o'
                % quote_plus(self.bucket))
        r = requests.post(path,
                          params={'uploadType': 'media', 'name': self.key},
                          headers=head, data=data)
        validate_response(r, path)

    def _fetch(self, start, end):
        # force read to 5MB boundaries
        start = start // (5 * 2**20) * 5 * 2**20
        end = (end // (5 * 2 ** 20) + 1) * 5 * 2 ** 20
        if self.start is None and self.end is None:
            # First read
            self.start = start
            self.end = end + self.blocksize
            self.cache = _fetch_range(self.gcsfs.header, self.details,
                                      start, self.end)
        if start < self.start:
            new = _fetch_range(self.gcsfs.header, self.details,
                               start, self.start)
            self.start = start
            self.cache = new + self.cache
        if end > self.end:
            if self.end > self.size:
                return
            new = _fetch_range(self.gcsfs.header, self.details,
                               self.end, end + self.blocksize)
            self.end = end + self.blocksize
            self.cache = self.cache + new

    def read(self, length=-1):
        """
        Return data from cache, or fetch pieces as necessary

        Parameters
        ----------
        length : int (-1)
            Number of bytes to read; if <0, all remaining bytes.
        """
        if self.mode != 'rb':
            raise ValueError('File not in read mode')
        if length < 0:
            length = self.size
        if self.closed:
            raise ValueError('I/O operation on closed file.')
        self._fetch(self.loc, self.loc + length)
        out = self.cache[self.loc - self.start:
                         self.loc - self.start + length]
        self.loc += len(out)
        if self.trim:
            num = (self.loc - self.start) // self.blocksize - 1
            if num > 0:
                self.start += self.blocksize * num
                self.cache = self.cache[self.blocksize * num:]
        return out

    def close(self):
        """ Close file """
        if self.closed:
            return
        if self.mode == 'rb':
            self.cache = None
        else:
            self.flush(force=True)
            self.gcsfs.invalidate_cache(self.bucket)
        self.closed = True

    def readable(self):
        """Return whether the GCSFile was opened for reading"""
        return self.mode == 'rb'

    def seekable(self):
        """Return whether the GCSFile is seekable (only in read mode)"""
        return self.readable()

    def writable(self):
        """Return whether the GCSFile was opened for writing"""
        return self.mode in {'wb', 'ab'}

    def __del__(self):
        self.close()

    def __str__(self):
        return "<GCSFile %s/%s>" % (self.bucket, self.key)

    __repr__ = __str__

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


def _fetch_range(head, obj_dict, start=None, end=None):
    """ Get data from GCS

    head : dict
        Contains authorization header
    obj_dict : an entry from ls() or info()
    start, end : None or integers
        if not both None, fetch only given range
    """
    if DEBUG:
        print('Fetch: ', start, end)
    logger.debug("Fetch: {}/{}, {}-{}", obj_dict['name'], start, end)
    if start is not None or end is not None:
        start = start or 0
        end = end or 0
        head = head.copy()
        head['Range'] = 'bytes=%i-%i' % (start, end - 1)
    back = requests.get(obj_dict['mediaLink'], headers=head)
    data = back.content
    if data == b'Request range not satisfiable':
        return b''
    return data


def put_object(credentials, bucket, name, data):
    """ Simple put, up to 5MB of data

    credentials : from auth()
    bucket : string
    name : object name
    data : binary
    """
    out = requests.post('https://www.googleapis.com/upload/storage/'
                        'v1/b/%s/o?uploadType=media&name=%s' % (
                            quote_plus(bucket), quote_plus(name)),
                        headers={'Authorization': 'Bearer ' +
                                                  credentials.access_token,
                                 'Content-Type': 'application/octet-stream',
                                 'Content-Length': len(data)}, data=data)
    assert out.status_code == 200


def ensure_writable(b):
    if PY2 and isinstance(b, array.array):
        return b.tostring()
    return b




import os
import gcsfs.core

RECORD_MODE = os.environ.get('GCSFS_RECORD_MODE', 'none')
TEST_PROJECT = os.environ.get('GCSFS_TEST_PROJECT', 'test_project')

TEST_BUCKET = 'gcsfs-testing'

FAKE_TOKEN = {'access_token': 'xxx', 'expires_in': 0,
              'grant_type': 'refresh_token',
              'refresh_token': 'xxx', 'timestamp': 1487859400.}

FAKE_TOKEN.update(gcsfs.core.not_secret)

FAKE_GOOGLE_TOKEN = {
  "client_id": "764086051850-6qr4p6gpi6hn506pt8ejuq83di341hur."
               "apps.googleusercontent.com",
  "client_secret": "d-FL95Q19q7MQmFpd7hHD0Ty",
  "refresh_token": "xxx",
  "type": "authorized_user"
}
GOOGLE_TOKEN = os.environ.get('GCSFS_GOOGLE_TOKEN', FAKE_GOOGLE_TOKEN)



from .core import GCSFileSystem


def register():
    """
    Make GCS filesystem available to dask parallel I/O.
    URLs can look like ``gs://myproject@mybucket/mypath.*.csv``

    Returns
    -------

    """
    global DaskGCSFileSystem
    import dask.bytes.core

    class DaskGCSFileSystem(GCSFileSystem, dask.bytes.core.FileSystem):

        sep = '/'

        def __init__(self, username=None, password=None, project=None,
                     path=None, host=None, **kwargs):
            if project is not None:
                if username is not None:
                    raise ValueError('GCS project defined twice: %s'
                                     % ((username, project),))
                username = project
            if username is not None:
                GCSFileSystem.__init__(self, project=username, **kwargs)
            else:
                GCSFileSystem.__init__(self, **kwargs)

        def open(self, path, mode='rb', **kwargs):
            bucket = kwargs.pop('host', '')
            gcs_path = bucket + path
            return GCSFileSystem.open(self, gcs_path, mode=mode)

        def glob(self, path, **kwargs):
            bucket = kwargs.pop('host', '')
            gcs_path = bucket + path
            return GCSFileSystem.glob(self, gcs_path)

        def mkdirs(self, path):
            pass  # no need to pre-make paths on GCS

        def ukey(self, path):
            return self.info(path)['etag']

        def size(self, path):
            return self.info(path)['size']

    dask.bytes.core._filesystems['gcs'] = DaskGCSFileSystem
    dask.bytes.core._filesystems['gs'] = DaskGCSFileSystem

try:
    register()
except ImportError as e:
    print(e)




import io
from itertools import chain
import os
import pytest

from gcsfs.tests.settings import TEST_PROJECT, GOOGLE_TOKEN, TEST_BUCKET
from gcsfs.tests.utils import (tempdir, token_restore, my_vcr, gcs_maker,
                               files, csv_files, text_files, a, b, c, d,
                               tmpfile)
from gcsfs.core import GCSFileSystem, quote_plus
from gcsfs.utils import seek_delimiter


@my_vcr.use_cassette(match=['all'])
def test_simple(token_restore):
    assert not GCSFileSystem.tokens
    gcs = GCSFileSystem(TEST_PROJECT, token=GOOGLE_TOKEN)
    assert gcs.ls('')

    # token is now cached
    gcs = GCSFileSystem(TEST_PROJECT)
    assert gcs.ls('')


@my_vcr.use_cassette(match=['all'])
def test_simple_upload(token_restore):
    with gcs_maker() as gcs:
        fn = TEST_BUCKET + '/test'
        with gcs.open(fn, 'wb') as f:
            f.write(b'zz')
        assert gcs.cat(fn) == b'zz'


@my_vcr.use_cassette(match=['all'])
def test_multi_upload(token_restore):
    with gcs_maker() as gcs:
        fn = TEST_BUCKET + '/test'
        d = b'01234567' * 2**15

        # something to write on close
        with gcs.open(fn, 'wb', block_size=2**18) as f:
            f.write(d)
            f.write(b'xx')
        assert gcs.cat(fn) == d + b'xx'

        # empty buffer on close
        with gcs.open(fn, 'wb', block_size=2**19) as f:
            f.write(d)
            f.write(b'xx')
            f.write(d)
        assert gcs.cat(fn) == d + b'xx' + d


@my_vcr.use_cassette(match=['all'])
def test_info(token_restore):
    with gcs_maker() as gcs:
        gcs.touch(a)
        assert gcs.info(a) == gcs.ls(a, detail=True)[0]


@my_vcr.use_cassette(match=['all'])
def test_ls2(token_restore):
    with gcs_maker() as gcs:
        assert TEST_BUCKET in gcs.ls('')
        with pytest.raises((OSError, IOError)):
            gcs.ls('nonexistent')
        fn = TEST_BUCKET+'/test/accounts.1.json'
        gcs.touch(fn)
        assert fn in gcs.ls(TEST_BUCKET+'/test')

@my_vcr.use_cassette(match=['all'])
def test_list_bucket_multipage(token_restore):
    with gcs_maker() as gcs:
        gcs.touch(a)
        gcs.touch(b)
        gcs.touch(c)
        dirs=gcs._list_bucket(TEST_BUCKET, max_results=2)
        assert len(dirs) == 3

@my_vcr.use_cassette(match=['all'])
def test_pickle(token_restore):
    with gcs_maker() as gcs:
        import pickle
        gcs2 = pickle.loads(pickle.dumps(gcs))

        # *values* may be equal during tests
        assert gcs.header is not gcs2.header
        gcs.touch(a)
        assert gcs.ls(TEST_BUCKET) == gcs2.ls(TEST_BUCKET)


@my_vcr.use_cassette(match=['all'])
def test_ls_touch(token_restore):
    with gcs_maker() as gcs:
        assert not gcs.exists(TEST_BUCKET+'/tmp/test')
        gcs.touch(a)
        gcs.touch(b)
        L = gcs.ls(TEST_BUCKET+'/tmp/test', True)
        assert set(d['name'] for d in L) == set([a, b])
        L = gcs.ls(TEST_BUCKET+'/tmp/test', False)
        assert set(L) == set([a, b])


@my_vcr.use_cassette(match=['all'])
def test_rm(token_restore):
    with gcs_maker() as gcs:
        assert not gcs.exists(a)
        gcs.touch(a)
        assert gcs.exists(a)
        gcs.rm(a)
        assert not gcs.exists(a)
        with pytest.raises((OSError, IOError)):
            gcs.rm(TEST_BUCKET+'/nonexistent')
        with pytest.raises((OSError, IOError)):
            gcs.rm('nonexistent')


@my_vcr.use_cassette(match=['all'])
def test_file_access(token_restore):
    with gcs_maker() as gcs:
        fn = TEST_BUCKET+'/nested/file1'
        data = b'hello\n'
        with gcs.open(fn, 'wb') as f:
            f.write(data)
        assert gcs.cat(fn) == data
        assert gcs.head(fn, 3) == data[:3]
        assert gcs.tail(fn, 3) == data[-3:]
        assert gcs.tail(fn, 10000) == data


@my_vcr.use_cassette(match=['all'])
def test_file_info(token_restore):
    with gcs_maker() as gcs:
        fn = TEST_BUCKET+'/nested/file1'
        data = b'hello\n'
        with gcs.open(fn, 'wb') as f:
            f.write(data)
        assert fn in gcs.walk(TEST_BUCKET)
        assert gcs.exists(fn)
        assert not gcs.exists(fn+'another')
        assert gcs.info(fn)['size'] == len(data)
        with pytest.raises((OSError, IOError)):
            gcs.info(fn+'another')


@my_vcr.use_cassette(match=['all'])
def test_du(token_restore):
    with gcs_maker(True) as gcs:
        d = gcs.du(TEST_BUCKET, deep=True)
        assert all(isinstance(v, int) and v >= 0 for v in d.values())
        assert TEST_BUCKET+'/nested/file1' in d

        assert gcs.du(TEST_BUCKET + '/test/', total=True) == sum(
                map(len, files.values()))
        assert gcs.du(TEST_BUCKET) == gcs.du('gcs://'+TEST_BUCKET)


@my_vcr.use_cassette(match=['all'])
def test_ls(token_restore):
    with gcs_maker(True) as gcs:
        fn = TEST_BUCKET+'/nested/file1'
        assert fn not in gcs.ls(TEST_BUCKET+'/')
        assert fn in gcs.ls(TEST_BUCKET+'/nested/')
        assert fn in gcs.ls(TEST_BUCKET+'/nested')
        assert gcs.ls('gcs://'+TEST_BUCKET+'/nested/') == gcs.ls(
                TEST_BUCKET+'/nested')


@my_vcr.use_cassette(match=['all'])
def test_ls_detail(token_restore):
    with gcs_maker(True) as gcs:
        L = gcs.ls(TEST_BUCKET+'/nested', detail=True)
        assert all(isinstance(item, dict) for item in L)


@my_vcr.use_cassette(match=['all'])
def test_gcs_glob(token_restore):
    with gcs_maker(True) as gcs:
        fn = TEST_BUCKET+'/nested/file1'
        assert fn not in gcs.glob(TEST_BUCKET+'/')
        assert fn not in gcs.glob(TEST_BUCKET+'/*')
        assert fn in gcs.glob(TEST_BUCKET+'/nested')
        assert fn in gcs.glob(TEST_BUCKET+'/nested/*')
        assert fn in gcs.glob(TEST_BUCKET+'/nested/file*')
        assert fn in gcs.glob(TEST_BUCKET+'/*/*')
        assert all(f in gcs.walk(TEST_BUCKET) for f in
                   gcs.glob(TEST_BUCKET+'/nested/*'))
        with pytest.raises(ValueError):
            gcs.glob('*')


@my_vcr.use_cassette(match=['all'])
def test_read_keys_from_bucket(token_restore):
    with gcs_maker(True) as gcs:
        for k, data in files.items():
            file_contents = gcs.cat('/'.join([TEST_BUCKET, k]))
            assert file_contents == data

        assert all(gcs.cat('/'.join([TEST_BUCKET, k])) ==
                   gcs.cat('gcs://' + '/'.join([TEST_BUCKET, k]))
                   for k in files)


@my_vcr.use_cassette(match=['all'])
def test_url(token_restore):
    with gcs_maker(True) as gcs:
        fn = TEST_BUCKET+'/nested/file1'
        url = gcs.url(fn)
        assert 'http' in url
        assert quote_plus('nested/file1') in url
        with gcs.open(fn) as f:
            assert 'http' in f.url()


@my_vcr.use_cassette(match=['all'])
def test_seek(token_restore):
    with gcs_maker(True) as gcs:
        with gcs.open(a, 'wb') as f:
            f.write(b'123')

        with gcs.open(a) as f:
            f.seek(1000)
            with pytest.raises(ValueError):
                f.seek(-1)
            with pytest.raises(ValueError):
                f.seek(-5, 2)
            with pytest.raises(ValueError):
                f.seek(0, 10)
            f.seek(0)
            assert f.read(1) == b'1'
            f.seek(0)
            assert f.read(1) == b'1'
            f.seek(3)
            assert f.read(1) == b''
            f.seek(-1, 2)
            assert f.read(1) == b'3'
            f.seek(-1, 1)
            f.seek(-1, 1)
            assert f.read(1) == b'2'
            for i in range(4):
                assert f.seek(i) == i


@my_vcr.use_cassette(match=['all'])
def test_bad_open(token_restore):
    with gcs_maker() as gcs:
        with pytest.raises(IOError):
            gcs.open('')


@my_vcr.use_cassette(match=['all'])
def test_copy(token_restore):
    with gcs_maker(True) as gcs:
        fn = TEST_BUCKET+'/test/accounts.1.json'
        gcs.copy(fn, fn+'2')
        assert gcs.cat(fn) == gcs.cat(fn+'2')


@my_vcr.use_cassette(match=['all'])
def test_move(token_restore):
    with gcs_maker(True) as gcs:
        fn = TEST_BUCKET+'/test/accounts.1.json'
        data = gcs.cat(fn)
        gcs.mv(fn, fn+'2')
        assert gcs.cat(fn+'2') == data
        assert not gcs.exists(fn)


@my_vcr.use_cassette(match=['all'])
def test_get_put(token_restore):
    with gcs_maker(True) as gcs:
        with tmpfile() as fn:
            gcs.get(TEST_BUCKET+'/test/accounts.1.json', fn)
            data = files['test/accounts.1.json']
            assert open(fn, 'rb').read() == data
            gcs.put(fn, TEST_BUCKET+'/temp')
            assert gcs.du(TEST_BUCKET+'/temp')[
                       TEST_BUCKET+'/temp'] == len(data)
            assert gcs.cat(TEST_BUCKET+'/temp') == data


@my_vcr.use_cassette(match=['all'])
def test_errors(token_restore):
    with gcs_maker() as gcs:
        with pytest.raises((IOError, OSError)):
            gcs.open(TEST_BUCKET+'/tmp/test/shfoshf', 'rb')

        ## This is fine, no need for interleving directories on gcs
        #with pytest.raises((IOError, OSError)):
        #    gcs.touch('tmp/test/shfoshf/x')

        with pytest.raises((IOError, OSError)):
            gcs.rm(TEST_BUCKET+'/tmp/test/shfoshf/x')

        with pytest.raises((IOError, OSError)):
            gcs.mv(TEST_BUCKET+'/tmp/test/shfoshf/x', 'tmp/test/shfoshf/y')

        with pytest.raises((IOError, OSError)):
            gcs.open('x', 'rb')

        with pytest.raises(IOError):
            gcs.rm('unknown')

        with pytest.raises(ValueError):
            with gcs.open(TEST_BUCKET+'/temp', 'wb') as f:
                f.read()

        with pytest.raises(ValueError):
            f = gcs.open(TEST_BUCKET+'/temp', 'rb')
            f.close()
            f.read()

        with pytest.raises(ValueError) as e:
            gcs.mkdir('/')
            assert 'bucket' in str(e)

        with pytest.raises(ValueError):
            gcs.walk('')

        with pytest.raises(ValueError):
            gcs.walk('gcs://')


@my_vcr.use_cassette(match=['all'])
def test_read_small(token_restore):
    with gcs_maker(True) as gcs:
        fn = TEST_BUCKET+'/2014-01-01.csv'
        with gcs.open(fn, 'rb', block_size=10) as f:
            out = []
            while True:
                data = f.read(3)
                if data == b'':
                    break
                out.append(data)
            assert gcs.cat(fn) == b''.join(out)
            # cache drop
            assert len(f.cache) < len(out)


@my_vcr.use_cassette(match=['all'])
def test_seek_delimiter(token_restore):
    with gcs_maker(True) as gcs:
        fn = 'test/accounts.1.json'
        data = files[fn]
        with gcs.open('/'.join([TEST_BUCKET, fn])) as f:
            seek_delimiter(f, b'}', 0)
            assert f.tell() == 0
            f.seek(1)
            seek_delimiter(f, b'}', 5)
            assert f.tell() == data.index(b'}') + 1
            seek_delimiter(f, b'\n', 5)
            assert f.tell() == data.index(b'\n') + 1
            f.seek(1, 1)
            ind = data.index(b'\n') + data[data.index(b'\n')+1:].index(b'\n') + 1
            seek_delimiter(f, b'\n', 5)
            assert f.tell() == ind + 1


@my_vcr.use_cassette(match=['all'])
def test_read_block(token_restore):
    with gcs_maker(True) as gcs:
        data = files['test/accounts.1.json']
        lines = io.BytesIO(data).readlines()
        path = TEST_BUCKET+'/test/accounts.1.json'
        assert gcs.read_block(path, 1, 35, b'\n') == lines[1]
        assert gcs.read_block(path, 0, 30, b'\n') == lines[0]
        assert gcs.read_block(path, 0, 35, b'\n') == lines[0] + lines[1]
        out = gcs.read_block(path, 0, 5000, b'\n')
        assert gcs.read_block(path, 0, 5000, b'\n') == data
        assert len(gcs.read_block(path, 0, 5)) == 5
        assert len(gcs.read_block(path, 4, 5000)) == len(data) - 4
        assert gcs.read_block(path, 5000, 5010) == b''

        assert gcs.read_block(path, 5, None) == gcs.read_block(path, 5, 1000)


@my_vcr.use_cassette(match=['all'])
def test_write_fails(token_restore):
    with gcs_maker() as gcs:
        with pytest.raises(NotImplementedError):
            gcs.open(TEST_BUCKET+'/temp', 'w')
        with pytest.raises(ValueError):
            gcs.touch(TEST_BUCKET+'/temp')
            gcs.open(TEST_BUCKET+'/temp', 'rb').write(b'hello')
        with pytest.raises(ValueError):
            with gcs.open(TEST_BUCKET+'/temp', 'wb') as f:
                f.write(b'hello')
                f.flush(force=True)
                f.write(b'world')
        f = gcs.open(TEST_BUCKET+'/temp', 'wb')
        f.close()
        with pytest.raises(ValueError):
            f.write(b'hello')
        with pytest.raises((OSError, IOError)):
            gcs.open('nonexistentbucket/temp', 'wb').close()


@my_vcr.use_cassette(match=['all'])
def test_write_blocks(token_restore):
    with gcs_maker() as gcs:
        with gcs.open(TEST_BUCKET+'/temp', 'wb', block_size=2**18) as f:
            f.write(b'a' * 100000)
            assert f.buffer.tell() == 100000
            assert not(f.offset)
            f.write(b'a' * 100000)
            f.write(b'a' * 100000)
            assert f.offset
        assert gcs.info(TEST_BUCKET+'/temp')['size'] == 300000
        with gcs.open(TEST_BUCKET+'/temp', 'wb', block_size=2**18) as f:
            f.write(b'a' * (2**18+1))
            # leftover bytes: GCS accepts blocks in multiples of 2**18 bytes
            assert f.buffer.tell() == 1
        assert gcs.info(TEST_BUCKET+'/temp')['size'] == 2**18+1


@my_vcr.use_cassette(match=['all'])
def test_readline(token_restore):
    with gcs_maker(True) as gcs:
        all_items = chain.from_iterable([
            files.items(), csv_files.items(), text_files.items()
        ])
        for k, data in all_items:
            with gcs.open('/'.join([TEST_BUCKET, k]), 'rb') as f:
                result = f.readline()
                expected = data.split(b'\n')[0] + (b'\n' if data.count(b'\n')
                                                   else b'')
            assert result == expected


@my_vcr.use_cassette(match=['all'])
def test_readline_from_cache(token_restore):
    with gcs_maker() as gcs:
        data = b'a,b\n11,22\n3,4'
        with gcs.open(a, 'wb') as f:
            f.write(data)

        with gcs.open(a, 'rb') as f:
            result = f.readline()
            assert result == b'a,b\n'
            assert f.loc == 4
            assert f.cache == data

            result = f.readline()
            assert result == b'11,22\n'
            assert f.loc == 10
            assert f.cache == data

            result = f.readline()
            assert result == b'3,4'
            assert f.loc == 13
            assert f.cache == data


@my_vcr.use_cassette(match=['all'])
def test_readline_partial(token_restore):
    with gcs_maker() as gcs:
        data = b'aaaaa,bbbbb\n12345,6789\n'
        with gcs.open(a, 'wb') as f:
            f.write(data)
        with gcs.open(a, 'rb') as f:
            result = f.readline(5)
            assert result == b'aaaaa'
            result = f.readline(5)
            assert result == b',bbbb'
            result = f.readline(5)
            assert result == b'b\n'
            result = f.readline()
            assert result == b'12345,6789\n'


@my_vcr.use_cassette(match=['all'])
def test_readline_empty(token_restore):
    with gcs_maker() as gcs:
        data = b''
        with gcs.open(a, 'wb') as f:
            f.write(data)
        with gcs.open(a, 'rb') as f:
            result = f.readline()
            assert result == data


@my_vcr.use_cassette(match=['all'])
def test_readline_blocksize(token_restore):
    with gcs_maker() as gcs:
        data = b'ab\n' + b'a' * (2**18) + b'\nab'
        with gcs.open(a, 'wb') as f:
            f.write(data)
        with gcs.open(a, 'rb', block_size=2**18) as f:
            result = f.readline()
            expected = b'ab\n'
            assert result == expected

            result = f.readline()
            expected = b'a' * (2**18) + b'\n'
            assert result == expected

            result = f.readline()
            expected = b'ab'
            assert result == expected


@my_vcr.use_cassette(match=['all'])
def test_next(token_restore):
    with gcs_maker(True) as gcs:
        expected = csv_files['2014-01-01.csv'].split(b'\n')[0] + b'\n'
        with gcs.open(TEST_BUCKET + '/2014-01-01.csv') as f:
            result = next(f)
            assert result == expected


@my_vcr.use_cassette(match=['all'])
def test_iterable(token_restore):
    with gcs_maker() as gcs:
        data = b'abc\n123'
        with gcs.open(a, 'wb') as f:
            f.write(data)
        with gcs.open(a) as f, io.BytesIO(data) as g:
            for fromgcs, fromio in zip(f, g):
                assert fromgcs == fromio
            f.seek(0)
            assert f.readline() == b'abc\n'
            assert f.readline() == b'123'
            f.seek(1)
            assert f.readline() == b'bc\n'
            assert f.readline(1) == b'1'
            assert f.readline() == b'23'

        with gcs.open(a) as f:
            out = list(f)
        with gcs.open(a) as f:
            out2 = f.readlines()
        assert out == out2
        assert b"".join(out) == data


@my_vcr.use_cassette(match=['all'])
def test_readable(token_restore):
    with gcs_maker() as gcs:
        with gcs.open(a, 'wb') as f:
            assert not f.readable()

        with gcs.open(a, 'rb') as f:
            assert f.readable()


@my_vcr.use_cassette(match=['all'])
def test_seekable(token_restore):
    with gcs_maker() as gcs:
        with gcs.open(a, 'wb') as f:
            assert not f.seekable()

        with gcs.open(a, 'rb') as f:
            assert f.seekable()


@my_vcr.use_cassette(match=['all'])
def test_writable(token_restore):
    with gcs_maker() as gcs:
        with gcs.open(a, 'wb') as f:
            assert f.writable()

        with gcs.open(a, 'rb') as f:
            assert not f.writable()


@my_vcr.use_cassette(match=['all'])
def test_merge(token_restore):
    with gcs_maker() as gcs:
        with gcs.open(a, 'wb') as f:
            f.write(b'a' * 100)

        with gcs.open(b, 'wb') as f:
            f.write(b'a' * 100)
        gcs.merge(TEST_BUCKET+'/joined', [a, b])
        assert gcs.info(TEST_BUCKET+'/joined')['size'] == 200


@my_vcr.use_cassette(match=['all'])
def test_bigger_than_block_read(token_restore):
    with gcs_maker(True) as gcs:
        with gcs.open(TEST_BUCKET+'/2014-01-01.csv', 'rb', block_size=3) as f:
            out = []
            while True:
                data = f.read(20)
                out.append(data)
                if len(data) == 0:
                    break
        assert b''.join(out) == csv_files['2014-01-01.csv']


@my_vcr.use_cassette(match=['all'])
def test_current(token_restore):
    with gcs_maker() as gcs:
        assert GCSFileSystem.current() is gcs


@my_vcr.use_cassette(match=['all'])
def test_array(token_restore):
    with gcs_maker() as gcs:
        from array import array
        data = array('B', [65] * 1000)

        with gcs.open(a, 'wb') as f:
            f.write(data)

        with gcs.open(a, 'rb') as f:
            out = f.read()
            assert out == b'A' * 1000
